@@ -1,6 +1,6 @@
 use std::{
     sync::{
-        atomic::{AtomicBool, AtomicUsize, Ordering},
+        atomic::{AtomicUsize, Ordering},
         Arc,
     },
     time::Instant,
@@ -41,8 +41,9 @@ pub(crate) async fn download_concurrently<C: CondowClient, R: Reporter>(
     downloader.download(ranges_stream).await
 }
 
+/// Spawns multiple [InternalDownloader]s to download parts
 struct ConcurrentDownloader<R: Reporter> {
-    downloaders: Vec<Downloader>,
+    downloaders: Vec<SequentialDownloader>,
     counter: usize,
     kill_switch: KillSwitch,
     config: Config,
@@ -63,13 +64,17 @@ impl<R: Reporter> ConcurrentDownloader<R> {
         let counter = Arc::new(AtomicUsize::new(0));
         let downloaders: Vec<_> = (0..n_concurrent)
             .map(|_| {
-                Downloader::new(
+                SequentialDownloader::new(
                     client.clone(),
-                    results_sender.clone(),
-                    kill_switch.clone(),
                     location.clone(),
                     config.buffer_size.into(),
-                    DownloadersWatcher::new(Arc::clone(&counter), reporter.clone(), started_at),
+                    DownloaderContext::new(
+                        results_sender.clone(),
+                        Arc::clone(&counter),
+                        kill_switch.clone(),
+                        reporter.clone(),
+                        started_at,
+                    ),
                 )
             })
             .collect();
@@ -123,84 +128,76 @@ impl<R: Reporter> ConcurrentDownloader<R> {
     }
 }
 
-struct Downloader {
-    sender: Sender<RangeRequest>,
-    kill_switch: KillSwitch,
+/// Downlads parts of download sequentially.
+///
+/// Takes the part ranges ([RangeRequest]s) of of a download and downloads them.
+///
+/// Spawns a task internally to process the parts to be downloaded one by one.
+/// The parts to be downloaded are enqueued in a channel.
+///
+/// Results are pushed into a channel via the [DownloaderContext].
+///
+/// Usually one `SequentialDownloader` is created for each level of
+/// concurrency.  
+struct SequentialDownloader {
+    request_sender: Sender<RangeRequest>,
 }
 
-impl Downloader {
+impl SequentialDownloader {
     pub fn new<C: CondowClient, R: Reporter>(
         client: ClientRetryWrapper<C>,
-        results_sender: UnboundedSender<ChunkStreamItem>,
-        kill_switch: KillSwitch,
         location: C::Location,
         buffer_size: usize,
-        watcher: DownloadersWatcher<R>,
+        mut context: DownloaderContext<R>,
     ) -> Self {
-        let (sender, request_receiver) = mpsc::channel::<RangeRequest>(buffer_size);
+        let (request_sender, request_receiver) = mpsc::channel::<RangeRequest>(buffer_size);
 
-        tokio::spawn({
-            let kill_switch = kill_switch.clone();
-            async move {
-                let mut request_receiver = Box::pin(request_receiver);
-                while let Some(range_request) = request_receiver.next().await {
-                    if kill_switch.is_pushed() {
-                        break;
-                    }
+        tokio::spawn(async move {
+            let mut request_receiver = Box::pin(request_receiver);
+            while let Some(range_request) = request_receiver.next().await {
+                if context.kill_switch.is_pushed() {
+                    // That failed task should have already sent an error...
+                    // ...but we do not want to prove that...
+                    context.send_err(CondowError::new_other(
+                        "another download task already failed",
+                    ));
+                    return;
+                }
 
-                    match client
-                        .download(
-                            location.clone(),
-                            DownloadSpec::Range(range_request.blob_range),
-                            &watcher.reporter,
-                        )
-                        .await
-                    {
-                        Ok((bytes_stream, _total_bytes)) => {
-                            if consume_and_dispatch_bytes(
-                                bytes_stream,
-                                &results_sender,
-                                range_request,
-                                watcher.reporter.clone(),
-                            )
+                match client
+                    .download(
+                        location.clone(),
+                        DownloadSpec::Range(range_request.blob_range),
+                        &context.reporter,
+                    )
+                    .await
+                {
+                    Ok((bytes_stream, _total_bytes)) => {
+                        if consume_and_dispatch_bytes(bytes_stream, &mut context, range_request)
                             .await
                             .is_err()
-                            {
-                                kill_switch.push_the_button();
-                                watcher.mark_failed();
-                                request_receiver.close();
-                                break;
-                            }
+                        {
+                            return;
                         }
-                        Err(err) => {
-                            kill_switch.push_the_button();
-                            watcher.mark_failed();
-                            request_receiver.close();
-                            let _ = results_sender.unbounded_send(Err(err));
-                            break;
-                        }
-                    };
-                }
-                drop(watcher);
+                    }
+                    Err(err) => {
+                        context.send_err(err);
+                        return;
+                    }
+                };
             }
+            context.mark_successful();
+            drop(context);
         });
 
-        Downloader {
-            sender,
-            kill_switch,
-        }
+        SequentialDownloader { request_sender }
     }
 
     pub fn enqueue(&mut self, req: RangeRequest) -> Result<Option<RangeRequest>, ()> {
-        if self.kill_switch.is_pushed() {
-            return Err(());
-        }
-
-        match self.sender.try_send(req) {
+        match self.request_sender.try_send(req) {
             Ok(()) => Ok(None),
             Err(err) => {
                 if err.is_disconnected() {
-                    self.kill_switch.push_the_button();
                     Err(())
                 } else {
                     Ok(Some(err.into_inner()))
@@ -210,34 +207,79 @@ impl Downloader {
     }
 }
 
-struct DownloadersWatcher<R: Reporter> {
+struct DownloaderContext<R: Reporter> {
     started_at: Instant,
     counter: Arc<AtomicUsize>,
-    is_failed: Arc<AtomicBool>,
+    kill_switch: KillSwitch,
     reporter: R,
+    results_sender: UnboundedSender<ChunkStreamItem>,
+    completed: bool,
 }
 
-impl<R: Reporter> DownloadersWatcher<R> {
-    pub fn new(counter: Arc<AtomicUsize>, reporter: R, started_at: Instant) -> Self {
+impl<R: Reporter> DownloaderContext<R> {
+    pub fn new(
+        results_sender: UnboundedSender<ChunkStreamItem>,
+        counter: Arc<AtomicUsize>,
+        kill_switch: KillSwitch,
+        reporter: R,
+        started_at: Instant,
+    ) -> Self {
         counter.fetch_add(1, Ordering::SeqCst);
         Self {
             counter,
             reporter,
-            is_failed: Arc::new(AtomicBool::new(false)),
+            kill_switch,
             started_at,
+            results_sender,
+            completed: false,
         }
     }
 
-    pub fn mark_failed(&self) {
-        self.is_failed.store(true, Ordering::SeqCst);
+    pub fn send_chunk(&self, chunk: Chunk) -> Result<(), ()> {
+        if self.results_sender.unbounded_send(Ok(chunk)).is_ok() {
+            return Ok(());
+        }
+
+        self.kill_switch.push_the_button();
+
+        return Err(());
+    }
+
+    /// Send an error and mark as completed
+    pub fn send_err(&mut self, err: CondowError) {
+        let _ = self.results_sender.unbounded_send(Err(err));
+        self.completed = true;
+        self.kill_switch.push_the_button();
+    }
+
+    /// Mark the download as complete if successful
+    ///
+    /// This must be called upon succesful termination of an [InternalDownloader].
+    ///
+    /// If the download was not marked complete, an error will be sent when dropped
+    /// (a panic is assumed).
+    pub fn mark_successful(&mut self) {
+        self.completed = true;
     }
 }
 
-impl<R: Reporter> Drop for DownloadersWatcher<R> {
+impl<R: Reporter> Drop for DownloaderContext<R> {
     fn drop(&mut self) {
+        if !self.completed {
+            self.kill_switch.push_the_button();
+
+            let err = if std::thread::panicking() {
+                self.reporter.panic_detected("panic detected in downloader");
+                CondowError::new_other("download ended unexpectedly due to a panic")
+            } else {
+                CondowError::new_other("download ended unexpectetly")
+            };
+            let _ = self.results_sender.unbounded_send(Err(err));
+        }
+
         self.counter.fetch_sub(1, Ordering::SeqCst);
         if self.counter.load(Ordering::SeqCst) == 0 {
-            if self.is_failed.load(Ordering::SeqCst) {
+            if self.kill_switch.is_pushed() {
                 self.reporter
                     .download_failed(Some(self.started_at.elapsed()))
             } else {
@@ -249,9 +291,8 @@ impl<R: Reporter> Drop for DownloadersWatcher<R> {
 
 async fn consume_and_dispatch_bytes<R: Reporter>(
     mut bytes_stream: BytesStream,
-    results_sender: &UnboundedSender<ChunkStreamItem>,
+    context: &mut DownloaderContext<R>,
     range_request: RangeRequest,
-    reporter: R,
 ) -> Result<(), ()> {
     let mut chunk_index = 0;
     let mut offset_in_range = 0;
@@ -260,7 +301,9 @@ async fn consume_and_dispatch_bytes<R: Reporter>(
     let part_start = Instant::now();
     let mut chunk_start = Instant::now();
 
-    reporter.part_started(range_request.part_index, range_request.blob_range);
+    context
+        .reporter
+        .part_started(range_request.part_index, range_request.blob_range);
 
     while let Some(bytes_res) = bytes_stream.next().await {
         match bytes_res {
@@ -271,41 +314,44 @@ async fn consume_and_dispatch_bytes<R: Reporter>(
                 bytes_received += bytes.len() as u64;
 
                 if bytes_received > bytes_expected {
-                    let msg = Err(CondowError::new_other(format!(
+                    let err = CondowError::new_other(format!(
                         "received more bytes than expected for part {} ({}..={}). expected {}, received {}",
                         range_request.part_index,
                         range_request.blob_range.start(),
                         range_request.blob_range.end_incl(),
                         range_request.blob_range.len(),
                         bytes_received
-                    )));
-                    let _ = results_sender.unbounded_send(msg);
+                    ));
+                    context.send_err(err);
                     return Err(());
                 }
 
-                reporter.chunk_completed(range_request.part_index, chunk_index, n_bytes, t_chunk);
+                context.reporter.chunk_completed(
+                    range_request.part_index,
+                    chunk_index,
+                    n_bytes,
+                    t_chunk,
+                );
 
-                results_sender
-                    .unbounded_send(Ok(Chunk {
-                        part_index: range_request.part_index,
-                        chunk_index,
-                        blob_offset: range_request.blob_range.start() + offset_in_range,
-                        range_offset: range_request.range_offset + offset_in_range,
-                        bytes,
-                        bytes_left: bytes_expected - bytes_received,
-                    }))
-                    .map_err(|_| ())?;
+                context.send_chunk(Chunk {
+                    part_index: range_request.part_index,
+                    chunk_index,
+                    blob_offset: range_request.blob_range.start() + offset_in_range,
+                    range_offset: range_request.range_offset + offset_in_range,
+                    bytes,
+                    bytes_left: bytes_expected - bytes_received,
+                })?;
                 chunk_index += 1;
                 offset_in_range += n_bytes as u64;
             }
             Err(IoError(msg)) => {
-                let _ = results_sender.unbounded_send(Err(CondowError::new_io(msg)));
+                context.send_err(CondowError::new_io(msg));
                 return Err(());
             }
         }
     }
 
-    reporter.part_completed(
+    context.reporter.part_completed(
         range_request.part_index,
         chunk_index,
         bytes_received,
@@ -313,23 +359,22 @@ async fn consume_and_dispatch_bytes<R: Reporter>(
     );
 
     if bytes_received != bytes_expected {
-        let msg = Err(CondowError::new_other(format!(
+        let err = CondowError::new_other(format!(
             "received wrong number of bytes for part {} ({}..={}). expected {}, received {}",
             range_request.part_index,
             range_request.blob_range.start(),
             range_request.blob_range.end_incl(),
             range_request.blob_range.len(),
             bytes_received
-        )));
-        let _ = results_sender.unbounded_send(msg);
+        ));
+        let _ = context.send_err(err);
         Err(())
     } else {
         Ok(())
     }
 }
-
 #[cfg(test)]
-mod tests {
+mod downloader_tests {
     use std::{
         sync::{atomic::AtomicUsize, Arc},
         time::Instant,
@@ -338,10 +383,13 @@ mod tests {
     use futures::StreamExt;
 
     use crate::{
-        condow_client::NoLocation,
+        condow_client::{
+            failing_client_simulator::FailingClientSimulatorBuilder, CondowClient, NoLocation,
+        },
         config::Config,
+        errors::{CondowError, CondowErrorKind},
         machinery::{
-            downloaders::{Downloader, DownloadersWatcher},
+            downloaders::{DownloaderContext, SequentialDownloader},
             range_stream::RangeStream,
             KillSwitch,
         },
@@ -360,11 +408,28 @@ mod tests {
             InclusiveRange(0, 9),
             InclusiveRange(0, 10),
         ] {
-            check(range, client.clone(), 10).await
+            check(range, client.clone(), 10).await.unwrap()
         }
     }
 
-    async fn check(range: InclusiveRange, client: TestCondowClient, part_size_bytes: u64) {
+    #[tokio::test]
+    async fn failing_request() {
+        let blob = (0u8..100).collect::<Vec<_>>();
+        let client = FailingClientSimulatorBuilder::default()
+            .blob(blob)
+            .chunk_size(100)
+            .responses()
+            .failure(CondowErrorKind::Other)
+            .finish();
+
+        assert!(check(InclusiveRange(0, 99), client, 100).await.is_err());
+    }
+
+    async fn check<C: CondowClient<Location = NoLocation>>(
+        range: InclusiveRange,
+        client: C,
+        part_size_bytes: u64,
+    ) -> Result<(), CondowError> {
         let config = Config::default()
             .buffer_size(10)
             .buffers_full_delay_ms(0)
@@ -378,13 +443,17 @@ mod tests {
 
         let (result_stream, results_sender) = ChunkStream::new(bytes_hint);
 
-        let mut downloader = Downloader::new(
+        let mut downloader = SequentialDownloader::new(
             client.into(),
-            results_sender,
-            KillSwitch::new(),
             NoLocation,
             config.buffer_size.into(),
-            DownloadersWatcher::new(Arc::new(AtomicUsize::new(0)), NoReporting, Instant::now()),
+            DownloaderContext::new(
+                results_sender,
+                Arc::new(AtomicUsize::new(0)),
+                KillSwitch::new(),
+                NoReporting,
+                Instant::now(),
+            ),
         );
 
         while let Some(next) = ranges_stream.next().await {
@@ -394,7 +463,7 @@ mod tests {
         drop(downloader); // Ends the stream
 
         let result = result_stream.collect::<Vec<_>>().await;
-        let result = result.into_iter().collect::<Result<Vec<_>, _>>().unwrap();
+        let result = result.into_iter().collect::<Result<Vec<_>, _>>()?;
 
         let total_bytes: u64 = result.iter().map(|c| c.bytes.len() as u64).sum();
         assert_eq!(total_bytes, range.len(), "total_bytes");
@@ -423,5 +492,7 @@ mod tests {
             next_range_offset += bytes.len() as u64;
             next_blob_offset += bytes.len() as u64;
         });
+
+        Ok(())
     }
 }
